/*
 * Raw to PPM image converter.
 * Intended for converting a Linux framebuffer to other formats.
 * Currently supports only 16-bit (5-6-5) raw images.
 * Copyright (C) 1999-2005 Scott Scriven
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Instructions for compiling:
 *   On most machines, this should work:
 *     cc -o raw2ppm raw2ppm.c
 */
#include <stdio.h>
#include <string.h>


int help();
int convert(char *infilename, char *outfilename, 
	    int w, int h, int bswap, int rotate);
unsigned char * load_image(char *infilename, int w, int h, int bswap);
int save_image(char *outfilename, unsigned char *image, int w, int h);
int rotate_image(unsigned char *image, int w, int h, int rotate);


////////////////////////////////////////////////////////////
//
int help()
{
   char * helptext = "Usage: raw2ppm [options] [infile]\n"
"Options include:\n"
"\t-h  \t--help     \tShow this help screen\n"
"\t-i f\t--in       \tInput filename [/dev/fb]\n"
"\t-o f\t--out      \tOutput filename [shot.ppm]\n"
"\t-s  \t--byte-swap\tUse this if the colors look wrong\n"
"\t-w n\t--width    \tSet image width [320]\n"
"\t-h n\t--height   \tSet image height [240]\n"
"\t-p  \t--portrait \tRotate image 270 degrees\n"
"\t-l  \t--landscape\tDo not rotate\n"
"";
   printf(helptext);
   exit(0);
}

////////////////////////////////////////////////////////////
//
int convert(char *infilename, char *outfilename, 
	    int w, int h, int bswap, int rotate)
{
  unsigned char *image;

  image = load_image(infilename, w, h, bswap);
  if(! image)
  {
     printf("Unable to load image: %s\n", infilename);
     return -1;
  }
  
  if(rotate)
  {
     int i;
     i = w;  w = h;  h = i;
     if(0 != rotate_image(image, w, h, rotate))
     {
	printf("Error rotating image.\n");
	return -1;
     }
  }
  
  if(0 != save_image(outfilename, image, w, h))
  {
     printf("Unable to save image: %s\n", outfilename);
     return -1;
  }

  return 0;
}

////////////////////////////////////////////////////////////
//
unsigned char * load_image(char *infilename, int w, int h, int bswap)
{
  FILE *in;
  int len;
  unsigned short *buf;
  unsigned char *image;
  int offset=0;

  image = (unsigned char *)malloc(w*h*3);
  if(! image)
  {
     printf("Unable to malloc image.\n");
     return NULL;
  }
  buf = (unsigned short *)malloc(sizeof(unsigned short)*w);
  if(! buf)
  {
     printf("Unable to malloc buf.\n");
     free(image);
     return NULL;
  }

  in = fopen(infilename, "rb");
  if(!in)
  {
     printf("Unable to open input file: %s\n", infilename);
     free(image);
     free(buf);
     return NULL;
  }


  /* fprintf(out, "P6\%d %d\n255\n", w, h); */

  while ((offset < w*h*3)  &&  (len = fread(buf, 2, w, in)) != 0) {
    int i;

    for (i = 0; i < len; i++) {
       unsigned char r, g, b;
       if(bswap) /* Swap the byte order */
       {
	  buf[i] = ((buf[i] & 0xff00) >> 8) + ((buf[i] & 0x00ff) << 8);
       }
       r = (buf[i] & 0xF800) >> 8;
       g = (buf[i] & 0x07E0) >> 3;
       b = (buf[i] & 0x001F) << 3;
       image[offset++] = r;
       image[offset++] = g;
       image[offset++] = b;
       /* fprintf(out, "%c%c%c", r, g, b); */
    }
  }
  fclose(in);
  free(buf);

  return image;
}

////////////////////////////////////////////////////////////
//
int save_image(char *outfilename, unsigned char *image, int w, int h)
{
  FILE *out;
  int i;
  unsigned char r, g, b;

  out = fopen(outfilename, "wb");
  if(!out)
  {
     printf("Unable to open output file: %s\n", outfilename);
     return -1;
  }


  fprintf(out, "P6\%d %d\n255\n", w, h);

  for (i = 0; i < w*h*3; ) {
     r = image[i++];
     g = image[i++];
     b = image[i++];
     fprintf(out, "%c%c%c", r, g, b);
  }
  fclose(out);

  return 0;
}


////////////////////////////////////////////////////////////
// Rotate image 270 degrees.
// TODO: support rotation to 90 and 180 degrees
//
int rotate_image(unsigned char *image, int w, int h, int rotate)
{
   unsigned char *copy;
   int o1, o2;
   int x1, y1, x2, y2;

   copy = (unsigned char *)malloc(w*h*3);
   if(! copy)
   {
      printf("Unable to malloc image copy.\n");
      return -1;
   }

   memcpy(copy, image, w*h*3);

   for(y1=0; y1<h; y1++)
   {
      o1 = y1 * w * 3;
      for(x1=0; x1<w; x1++)
      {
	 x2 = h-1-y1;
	 y2 = x1;
	 o2 = (y2*h*3) + (x2*3);
	 image[o1++] = copy[o2++];
	 image[o1++] = copy[o2++];
	 image[o1++] = copy[o2++];
      }
   }
   free(copy);

   return 0;
}

////////////////////////////////////////////////////////////
//
int main(int argc, char *argv[]) {
  int w, h;
  int bswap=0;
  char *infilename = "/dev/fb";
  char *outfilename = "shot.ppm";
  int i;
  int rotate=0;
  
  /* Leftover code from original program, which required X11
  Display *dpy = XOpenDisplay("");
  w = DisplayWidth(dpy, DefaultScreen(dpy));
  h = DisplayHeight(dpy, DefaultScreen(dpy));
  */

  w = 320;
  h = 240;

  for(i=1; i<argc; i++)
  {
     if(argv[i][0] == '-')
     {
	/* The "-h" option specifies height. 
	   But it will call help if no height is given.  :)  */
	if((0 == strcmp(argv[i], "--help")))
	{
	   help();
	}
	else if((0 == strcmp(argv[i], "-s")) ||
	   (0 == strcmp(argv[i], "--byte-swap")))
	{
	   bswap=1;
	}
	else if((0 == strcmp(argv[i], "-w")) ||
		(0 == strcmp(argv[i], "--width")))
	{
	   if(i + 1 >= argc) help();
	   w = atoi(argv[++i]);
	}
	else if((0 == strcmp(argv[i], "-h")) ||
		(0 == strcmp(argv[i], "--height")))
	{
	   if(i + 1 >= argc) help();
	   h = atoi(argv[++i]);
	}
	else if((0 == strcmp(argv[i], "-i")) ||
		(0 == strcmp(argv[i], "--in")))
	{
	   if(i + 1 >= argc) help();
	   infilename = argv[++i];
	}
	else if((0 == strcmp(argv[i], "-o")) ||
		(0 == strcmp(argv[i], "--out")))
	{
	   if(i + 1 >= argc) help();
	   outfilename = argv[++i];
	}
	else if((0 == strcmp(argv[i], "-p")) ||
		(0 == strcmp(argv[i], "--portrait")))
	{
	   rotate=1;
	}
	else if((0 == strcmp(argv[i], "-l")) ||
		(0 == strcmp(argv[i], "--landscape")))
	{
	   rotate=0;
	}
     }
     else /* parameter is an input filename */
     {
	infilename = argv[i];
     }
  }

  if(w < 1  ||  h < 1) help();
  
  return convert(infilename, outfilename, w, h, bswap, rotate);
}
